-----------------------------------------------------------------------------
--                   _        _____ _               _                      --
--                  / \      | ____| |__   ___ _ __| | ___                 --
--                 / _ \     |  _| | '_ \ / _ \ '__| |/ _ \                --
--                / ___ \ _  | |___| |_) |  __/ |  | |  __/                --
--               /_/   \_(_) |_____|_.__/ \___|_|  |_|\___|                --
--                                                                         --
-----------------------------------------------------------------------------
-- ID:                $Id: update.lua 19415 2022-04-27 14:21:41Z juergen.schade $
-- Last committed:    $Revision: 19415 $
-- Last changed by:   $Author: juergen.schade $
-- Last changed date: $Date: 2022-04-27 16:21:41 +0200 (Mi, 27 Apr 2022) $
-----------------------------------------------------------------------------

-----------------------------------------------------------------------------
--@file    update.lua
--@brief
--@author  Hittmeyer stefan.hittmeyer@a-eberle.de
--
--Projekt : System 2010\n
--Betreuer(in): Hittmeyer, A. Eberle GmbH & Co. KG \n
--
--@version $Revision: 19415 $
-----------------------------------------------------------------------------

------------------------------------------------------------------------------
-- Private typedefs, enums, unions, variables
------------------------------------------------------------------------------

-- Caller of this script should set/declare _PROPER globally before call
-- _PROPER == true -> cleanup and reset to default; _PROPER == false -> incremental update
-- in order to avoid incompatibility we check the declaration of USE_RPOPPER and declare it if not existing
if (nil == _PROPER)    then _PROPER = false end
-- true if the eventRec and rec log-files should be removed, else false
if (nil == _RM_LOG)    then _RM_LOG = true end
-- downgrading is disabled (in case of downgrading, the ips settings will be discarded)
if (nil == _DOWNGRADE) then _DOWNGRADE = false end

local NAME                 = "update.lua"
local VERSION              = "0.35"

local KOMPAKT_ARTICLE_NR   = 150010500
local SMART_ARTICLE_NR     = 1197501

local FW_PATH              = "/mnt/flashdrive0"
local FW_REV               = 0 -- nil is not a valid value comparable with a number

local SRC_PATH             = "/root"
local ARM_IMAGE            = "/app.ais"
local DSP_IMAGE            = "/dummy.bin"
local BOOTER_IMAGE         = "/booter.ais"

-- load library
local lib = require"struct"

-- Lua 5.1 x Lua 5.2
local unpack = unpack or table.unpack

------------------------------------------------------------------------------
-- Private Helper Function Definitions
------------------------------------------------------------------------------
------------------------------------------------------------------------------
--@fn      function getFwPath()
--@brief   creates the firmwarepath for Smart & Kompakt
--
--@return
--       \li success: path to the firmware-image-folder (SD | flashdrive)
--       \li error: empty string
------------------------------------------------------------------------------
function getFwPath()
   if (false == FW_PATH) then
      eeprom_raw = eeprom.read(0, 0x0a)
      eeprom_serial = table.pack(lib.unpack("iiii", eeprom_raw))

      if eeprom_serial[3] == KOMPAKT_ARTICLE_NR then
      FW_PATH = "/mnt/sd0"
      elseif eeprom_serial[3] == SMART_ARTICLE_NR then
         FW_PATH = "/mnt/flashdrive0"
      else FW_PATH = ""
      end
   end
   return FW_PATH
end

------------------------------------------------------------------------------
--@fn    uint32ToString(x)
--@brief function for converting an uint32 to a string
--
--@param x is the value to convert
------------------------------------------------------------------------------
function uint32ToString(x)
    local b1=x%256  x=(x-x%256)/256
    local b2=x%256  x=(x-x%256)/256
    local b3=x%256  x=(x-x%256)/256
    local b4=x%256  x=(x-x%256)/256
    return string.char(b1,b2,b3,b4)
end

------------------------------------------------------------------------------
--@fn      function updateFwFile(filePath, fileName, fwType)
--@brief   updates the firmware files (binaries)
--
--@param filePath
--       path the new files (from update *tgz)
--@param fileName
--       the name of the file to be updated
--@param fwType
--       type of the firmware file (ARM DSP BOOTER)
--@return
--       \li success: true
--       \li error:   false
------------------------------------------------------------------------------
function updateFwFile(filePath, fileName, fwType)
   local size = fs.du(filePath..fileName)
   local fwPath
   local oldMd5
   local newMd5

   if ("ARM" == fwType) then fwPath = "/mnt/flashdrive0/sys/images"
   else                      fwPath = getFwPath().."/sys/images" end

   -- check if a update is required!
   oldMd5 = uti.md5(fwPath..fileName)
   newMd5 = uti.md5(filePath..fileName)
   if (oldMd5 == newMd5) then
     sprint(string.format("%-5s %-7s image -> untouched %s... ","equal", fwType, fwPath..fileName))
     fs.rm(filePath..fileName)
     return true
   end

   if (size ~= nil and size > 0) then
      sprint(string.format("%-5s %-7s image -> updating %s...","found", fwType, fwPath..fileName))
      ret = fs.mv(fwPath..fileName, fwPath..fileName..".bak")
      if (ret) then return false end
      ret = fs.mv(filePath..fileName, fwPath..fileName)
      if (ret) then return false end
      if ("BOOTER" == fwType) then
         sprint(string.format("O.K.\n                    -> flashing %s...", fwPath..fileName))
         if (nil ~= fs.flashBooter(fwPath..fileName)) then return false end
      end
      return true
   else
      sprint(string.format("%-5s %-7s image found...","no", fwType))
      return true
   end
end

------------------------------------------------------------------------------
--@fn      function mvUpdFold(srcPath, dstPath)
--@brief   moves folder with all its contents (recursively)
--
--@param srcPath
--       source to be moved
--@param dstPath
--       destination to be moved to
--@return
--       \li success: true
--       \li error:   false
------------------------------------------------------------------------------
function mvUpdFold(srcPath, dstPath)
   fs.cd(srcPath)                      -- goto the source folder...
   local src = fs.ls()                       -- and get a listing of it
   for i,v in ipairs(src) do           -- for each element of the source folder listing...
      if src[i].type == "file" and src[i].name ~= "." and src[i].name ~= ".." then  -- if element is a file...
         --print(   srcPath..src[i].name, dstPath..src[i].name)
         fs.rm(dstPath..src[i].name)   -- remove the file within destination
         ret = fs.mv(srcPath..src[i].name, dstPath..src[i].name)  -- and move the source file to destination
         if (ret) then return false end   -- check if moving worked out and exit on error
      elseif src[i].type == "dir" and src[i].name ~= "." and src[i].name ~= ".." then -- else if the source obj is folder
         --print(srcPath..src[i].name.."/", dstPath..src[i].name.."/")
         local pwd = fs.pwd()
         if (nil ~= fs.cd(dstPath..src[i].name)) then fs.mkdir(dstPath..src[i].name) end
         fs.cd(pwd)
         ret = mvUpdFold(srcPath..src[i].name.."/", dstPath..src[i].name.."/") -- recursive call self with new path
         if (false == ret) then return false end   -- check if moving worked out and exit on error
      elseif src[i].type ~= "file" and src[i].type ~= "dir" then
         --print(string.format("Type \"%s\" is not supported", src[i].type))
         return false
      end
   end
   if (fs.rmdir(srcPath)) then return false end   -- check if moving worked out and exit on error
   return true
end

------------------------------------------------------------------------------
--@fn      function updateFolder(folderPath, folderName)
--@brief   updates a Folders content
--
--@param folderPath
--       path to the folder
--@param folderName
--       name of the folder within folderPath
--@return
--       \li success: true
--       \li error:   false
------------------------------------------------------------------------------
function updateFolder(folderPath, folderName)
   local dstPath = getFwPath()..folderName
   local srcPath = folderPath..folderName
   local size = fs.du(folderPath..folderName)
   local pwd = fs.pwd()

   if (nil ~= fs.cd(dstPath)) then fs.mkdir(dstPath) end -- create the destination folder if it doesn't exist
   fs.cd(pwd) -- go back to original path

   if (size ~= nil and size > 0) then
      sprint(string.format("%-5s %-7s image -> updating %s...","found", folderName, dstPath))
      ret = mvUpdFold(srcPath.."/", dstPath.."/")
      if (ret) then return true else return false end
   else
      sprint(string.format("%-5s %-7s image found...","no", folderName))
      return true
   end
end

------------------------------------------------------------------------------
--@fn      function makeFwPathPropper()
--@brief   cleans the filesystem for a fresh, factory-setup fw-update
------------------------------------------------------------------------------
function makeFwPathPropper()
   -- clear the FwPath by removing any file within FwPath
   fs.cd(getFwPath())                              -- delete any file within FwPath
   local src = fs.ls()                             -- get a listing of FwPath
   for i,v in ipairs(src) do                       -- for each element of the source folder listing...
      if src[i].type == "file" and src[i].name ~= "."
                               and src[i].name ~= ".." then
         fs.rm(src[i].name) end
   end

   -- clearing the /data folder while keeping our update-folder
   fs.cd(getFwPath().."/data")                              -- delete any file within FwPath
   local src = fs.ls()                             -- get a listing of FwPath
   -- delete all except the ccci, srb and eventRec folder and after that,
   -- we will call the functions for removing the files of srb and eventRec!
   for i,v in ipairs(src) do                       -- for each element of the source folder listing...
      if src[i].type == "file" and src[i].name ~= "."
                               and src[i].name ~= ".." then
         fs.rm(src[i].name)
      -- else if the source obj is folder and not the ccci/srb/eventRec folder
      elseif src[i].type == "dir" and src[i].name ~= "."
                                  and src[i].name ~= ".."
                                  and src[i].name ~= "ccci"
                                  and src[i].name ~= "srb"
                                  and src[i].name ~= "eventRec"
                                  and src[i].name ~= "licenseMan"
                                  and src[i].name ~= "usrMan" then
         fs.rmdir(src[i].name, "-r")
      end
   end
   -- remove the log files only if this is not a Enexis version
   if _RM_LOG then
      if 2280 > FW_REV then  -- ensure we are not crashing older system because SD-Sync accesses the data-section
         if nil == fs.cd("/mnt/sd0") then print("cleaning /data section elapsed due to mounted SD0!")
         else -- remove the folders (this is still a little dangerous)
            if rec.rmAllLog then rec.rmAllLog() else fs.rmdir("srb", "-r") end
            if evtrec.rmAllLog then evtrec.rmAllLog() else fs.rmdir("eventRec", "-r") end
         end
      else -- we can go the nice and easy way
         rec.rmAllLog()
         evtrec.rmAllLog()
      end
   end

   -- clearing config folder completely while keeping the config folder self
   fs.cd(getFwPath().."/config")                              -- delete any file within FwPath
   local src = fs.ls()                             -- get a listing of FwPath
   for i,v in ipairs(src) do                       -- for each element of the source folder listing...
      if src[i].type == "file" and src[i].name ~= "."
                               and src[i].name ~= ".." then
         fs.rm(src[i].name)

         -- else if the source obj is the usrMan folder which needs an other handling
      elseif src[i].type == "dir" and src[i].name == "usrMan" then
         -- is not nice to delete a list of files here
         -- update strategy for usrMan has to be developed anyway
         for x = 1,5 do fs.rm("usrMan/*.c*") end;
         fs.rm("usrMan/tallylog")

      -- else if the source obj is folder and not the ips folder
      -- IMPORTANT: if the _DOWNGRADE flag is set, the ips folder should also be deleted!
      elseif src[i].type == "dir" and src[i].name ~= "."
                                  and src[i].name ~= ".."
                                  and (src[i].name ~= "ips" or _DOWNGRADE == true) then
         fs.rmdir(src[i].name, "-r")
      end
   end

   -- remove the logd.log
   if(log) then log.rmLog() else uti.rmLog() end

   -- we can simply remove this folders
   fs.rmdir(getFwPath().."/testresults", "-r")     -- remove the booter-testresults
   fs.rmdir(getFwPath().."/final", "-r")           -- remove the factory finalize xml files
end

------------------------------------------------------------------------------
--@fn      cleanRoot(rootPath, errMsg)
--@brief   deletes update-files from update-root-folder
--
--@param rootPath
--       folder containing the update-files
--@param errMsg
--       errorMessage to be diplayed by assert, set to nil or false for no error
--       throwing
------------------------------------------------------------------------------
function cleanRoot(rootPath, errMsg)
   -- if errMsg is !false, this is an error-abbortion
   if (errMsg) then print"failed" end

   -- remove root folder
   if (fs.rmdir(rootPath..SRC_PATH, "-r")) then
      uti.unmute() assert(false, string.format("can't erase %s ", rootPath..SRC_PATH))
   end
   -- remove update.lua
   if (fs.rmdir(rootPath.."/"..NAME)) then
      uti.unmute() assert(false, string.format("can't erase %s ", rootPath.."/"..NAME))
   end
   -- remove the revLog_ARM.log
   if fs.rm(rootPath.."/".."revLog_ARM.log") then
      uti.unmute() assert(false, string.format("can't erase %s ", rootPath.."/".."revLog_ARM.log"))
   end
   -- remove the revLog_DSP.log
   if fs.rm(rootPath.."/".."revLog_DSP.log") then
      uti.unmute() assert(false, string.format("can't erase %s ", rootPath.."/".."revLog_DSP.log"))
   end
   uti.unmute()
   -- if errMsg is !false, this is an error-abbortion -> print the text within assertion
   if (errMsg) then assert(false, errMsg) end
end

------------------------------------------------------------------------------
--@fn      upgradingConfig()
--@brief   if config/tctl exists, moves config/tctl to config/iec870 and removes
--         the config/iec850
------------------------------------------------------------------------------
function upgradingConfig(fwPath)
  local cwd = fs.pwd()
  -- if tctl is existing -> rename it ...
  if nil == fs.cd(fwPath.."/config/tctl") then
  print("\nUpgrading SCADA configuration required!")
    fs.cd("/")
    print(string.format("  > moving   \"%s/config/tctl\" to \"%s/config/iec870\"", fwPath, fwPath))
    if fs.mv(fwPath.."/config/tctl", fwPath.."/config/iec870") then
      uti.unmute() assert(false, string.format("can't move \"%s/config/tctl\"", fwPath))
    end
    -- only if tctl was existing, remove also the iec850
    if nil == fs.cd(fwPath.."/config/iec850") then
      fs.cd("/")
      print(string.format("  > removing \"%s/config/iec850\"", fwPath))
      if fs.rmdir(fwPath.."/config/iec850", "-r") then
        uti.unmute() assert(false, string.format("can't remove \"%s/config/iec850\"", fwPath))
      end
    end
  end
  fs.cd(cwd)
end

-- ##############################################################################################################
-- ##############################################################################################################
-- ##############################################################################################################
------------------------------------------------------------------------------
--@fn      Public Main Script
--@brief   main of the update script
--
--@param cwd
--       current working directory containing the extracted update sources
------------------------------------------------------------------------------
-- preparations
print(string.format("%20s v%s on %s\n", NAME, VERSION, _VERSION))
local arg = table.pack(...)      -- this are the parameters given on call
assert(1 == arg.n, "need cwd!")  -- check the number of parameters first
uti.mute()                       -- mute the fs-lib (dbg,error,result prints are suppressed)

------------------------------------------------------------------------------
-- file-system safety; ensure there is enough space left to proceed
------------------------------------------------------------------------------
local used = fs.du("/mnt/flashdrive0")
uti.unmute()
assert(used < 975175680, "Not enough memory on flashdrive! Abort!")
uti.mute()

------------------------------------------------------------------------------
-- if the function for receiving the Firmware Revision exists, call it and save the
-- revision else the FW_REV will stay '0' (invalid)
-- uti.getRev was introduced with FW_REV == 1939
------------------------------------------------------------------------------
if uti.getRev then FW_REV = uti.getRev() end
uti.unmute()
print(string.format("Currently running firmware revision is %d", FW_REV))
uti.mute()                      

------------------------------------------------------------------------------
-- force abort in case of incremental update, if the current version is < 2385
------------------------------------------------------------------------------
if (_PROPER == false) then
   uti.unmute()
   assert(FW_REV >= 2380, string.format("Current Firmware %d is too old! Please execute a factory Firmware-Update", FW_REV))
   uti.mute()                      
end
------------------------------------------------------------------------------
-- do the whole Firmware update stuff
------------------------------------------------------------------------------
if (updateFwFile(arg[1]..SRC_PATH.."/sys/images", ARM_IMAGE, "ARM")) then print"O.K."
else cleanRoot(arg[1], "ARM image failed") end
if (updateFwFile(arg[1]..SRC_PATH.."/sys/images", DSP_IMAGE, "DSP")) then print"O.K."
else cleanRoot(arg[1], "DSP image failed") end
if (updateFwFile(arg[1]..SRC_PATH.."/sys/images", BOOTER_IMAGE, "BOOTER")) then print"O.K."
else cleanRoot(arg[1], "BOOTER image failed") end
fs.rm(arg[1]..SRC_PATH.."/sys/images")

------------------------------------------------------------------------------
-- cleaup the flash only if desired
------------------------------------------------------------------------------
if _PROPER then
   -- make the flashdrive propper
   sprint(string.format("Making %s propper ...", getFwPath()))
   makeFwPathPropper()
   print"O.K."
end

------------------------------------------------------------------------------
-- updates from Firmware Version 1.4.x and older to 1.6.x and newer
-- require moving the config/tctl folder to config/iec870 (swMod renamed)
-- and wiping the config/iec850 (feature was introduced with 1.6)
-- we have to do so before updateFolder is called for /config, else
-- the new iec850 will be removed forever
------------------------------------------------------------------------------
upgradingConfig(getFwPath())

------------------------------------------------------------------------------
-- do the whole flash-file-structure stuff
------------------------------------------------------------------------------
if (updateFolder(arg[1]..SRC_PATH, "/final")) then print"O.K."
else cleanRoot(arg[1], "/final failed") end
if (updateFolder(arg[1]..SRC_PATH, "/config")) then print"O.K."
else cleanRoot(arg[1], "/config failed") end
if (updateFolder(arg[1]..SRC_PATH, "/data")) then print"O.K."
else cleanRoot(arg[1], "/data failed") end
if (updateFolder(arg[1]..SRC_PATH, "/sys")) then print"O.K."
else cleanRoot(arg[1], "/sys failed") end

------------------------------------------------------------------------------
-- change the first retransmission tmo and the last retransmission tmo values
-- only for those firmwares, it is required, else skip this section
-- first occurrence of RTM_TMO hack was with        FW_REV == 1925
-- first update-package including this hack was     RW_REV == 1998
-- FW_REV == 2217 will include this hack so we can use this version as threshold
------------------------------------------------------------------------------
if 2117 > FW_REV then
   -- the values ar changed in the config.cfg and also in the default.cfg of ips module
   print"Changing TCP/IP default retransmission timeouts to 1s and 2.5s"
   local file = io.open("/mnt/flashdrive0/config/ips/default.cfg", "r+b")
   file:seek("set", "28")
   file:write(uint32ToString(0x3E8))
   file:write(uint32ToString(0x9C4))
   file:close()
   print"Changing TCP/IP current retransmission timeouts to 1s and 2.5s"
   file = io.open("/mnt/flashdrive0/config/ips/config.cfg", "r+b")
   file:seek("set", "28")
   file:write(uint32ToString(0x3E8))
   file:write(uint32ToString(0x9C4))
   file:close()
end

------------------------------------------------------------------------------
-- create the links to srb and evt if possible / necessary
------------------------------------------------------------------------------
if fs.ln and nil == fs.cd(getFwPath().."/data/www") then
   if fs.cd(getFwPath().."/data/www/recorder") then fs.ln(getFwPath().."/data/www/recorder", "/sys/schreiber") end
   if fs.cd(getFwPath().."/data/www/eventRec") then fs.ln(getFwPath().."/data/www/eventRec", "/sys/eventRec") end
end

------------------------------------------------------------------------------
-- cleanup the filesystem
------------------------------------------------------------------------------
cleanRoot(arg[1])

------------------------------------------------------------------------------
-- Success
------------------------------------------------------------------------------
print(string.format("End of %s v%s\n", NAME, VERSION))
uti.unmute()

------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- END OF SCRIPT --
